"""Controls the output."""

import logging
import sys
from typing import Any, Callable

from rich.console import Console
from rich.logging import RichHandler

console = Console()


FORMAT = '%(message)s'
logging.basicConfig(
    level='NOTSET',
    format=FORMAT,
    datefmt='[%X]',
    handlers=[RichHandler(rich_tracebacks=True)],
)

log = logging.getLogger('rich')


def exit_with_error(message: str):
    """Show a error message and finish execution.

    Args:
        message (str): Error message.
    """
    log.error(message, extra={'markup': True})
    sys.exit(1)


def run_task(
    running_message: str,
    success_message: str,
    task_function: Callable,
    *args: Any,
    **kwargs: Any,
) -> Any:
    """[summary].

    Args:
        running_message (str): [description]
        success_message (str): [description]
        task_function (Callable): [description]
        args (tuple): [description]
        kwargs (dict): [description]

    Returns:
        Any: [description]
    """
    with console.status(running_message):
        task_result = task_function(*args, **kwargs)

    if task_result[0]:
        exit_with_error(task_result[0])
    else:
        log.info(success_message, extra={'markup': True})

    if len(task_result) > 1:
        return task_result[1]



"""Run the Cakemix."""

import importlib
import sys
from typing import Any

if sys.version_info.minor >= 8:
    module_name = 'importlib.metadata'
else:
    module_name = 'importlib_metadata'

importlib_metadata: Any = importlib.import_module(module_name)

cakemix = list(
    filter(
        lambda entry_point: entry_point.name == 'cakemix',
        importlib_metadata.entry_points()['console_scripts'],
    ),
)[0]

cakemix.load()()



"""Manage database."""

from pathlib import Path
from typing import Any, Callable

from sqlalchemy import Column, ForeignKey, Integer, String, create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship, sessionmaker
from sqlalchemy_utils.types.choice import ChoiceType

Base = declarative_base()


class CakemixTable(Base):
    """Table with the information of the boilerplates."""

    __tablename__ = 'cakemixes'

    name = Column(String, primary_key=True)
    structure = Column(String, nullable=False)

    arguments = relationship('ArgumentTable')
    paths = relationship('PathTable')


class Cakemix(CakemixTable):
    """Provides high-level functions to manipulate the cakemixes."""

    def __init__(self, database: 'Database', **args):
        """Create a BoilerplateTable object.

        Args:
            database (Database): [description]
            args (tuple): [description]
        """
        super().__init__(**args)
        self._database = database

    def add_argument(self, **kwargs):
        """[summary].

        Args:
            kwargs (dict): [description]
        """
        self.arguments.append(
            self._database.add(
                self._database.argument_object, **kwargs,
            ),
        )

    def add_path(self, **kwargs):
        """[summary].

        Args:
            kwargs (dict): [description]
        """
        self.paths.append(
            self._database.add(
                self._database.path_object, **kwargs,
            ),
        )


class ArgumentTable(Base):
    """Table with boilerplate arguments."""

    parameter_type_choices = [
        ('argument', 'Argument'),
        ('option', 'Option'),
    ]

    input_type_choices = [
        ('text', 'Text'),
    ]

    only_choices = [
        ('argument', 'Argument'),
        ('question', 'Question'),
    ]

    __tablename__ = 'arguments'

    pk = Column(Integer, primary_key=True)
    cakemix_name = Column(
        Integer, ForeignKey('cakemixes.name'), nullable=False,
    )
    name = Column(String, nullable=False)
    parameter_type = Column(
        ChoiceType(
            parameter_type_choices,
        ), default='argument',
    )
    input_type = Column(ChoiceType(input_type_choices), default='text')
    only = Column(ChoiceType(only_choices))

    cakemix = relationship('Cakemix')


class PathTable(Base):
    """Table with the boilerplates paths."""

    content_type_choices = [
        ('plain_text', 'Plain text'),
        ('not_plain_text', 'Not plain text'),
        ('directory', 'Directory'),
    ]

    __tablename__ = 'paths'

    pk = Column(Integer, primary_key=True)
    cakemix_name = Column(
        Integer, ForeignKey('cakemixes.name'), nullable=False,
    )
    path = Column(String, nullable=False)
    content_type = Column(
        ChoiceType(
            content_type_choices,
        ), default='plain_text',
    )

    cakemix = relationship('Cakemix')


class Database(object):
    """Provides high-level functions to manipulate the database."""

    def __init__(self, in_memory: bool = False):
        """Define the database path.

        Args:
            in_memory (bool, optional): The database path. Defaults to False.
        """
        self.in_memory = in_memory

    def __enter__(self) -> 'Database':
        """Create a SQLAlchemy session.

        Returns:
            Database: [description]
        """
        if self.in_memory:
            database_url = 'sqlite://'
        else:
            database_path = Path.home() / '.cakemix' / 'database.sqlite'
            database_url = f'sqlite:///{database_path}'

        engine = create_engine(database_url, pool_pre_ping=True)

        Base.metadata.create_all(engine)

        self._session = sessionmaker(bind=engine)()

        self.cakemix_object = Cakemix
        self.argument_object = ArgumentTable
        self.path_object = PathTable

        return self

    def __exit__(self, exception_type, exception_value, traceback):
        """Close the SQLAlchemy session.

        Args:
            exception_type ([type]): [description]
            exception_value ([type]): [description]
            traceback ([type]): [description]
        """
        self._session.close()

    def add(self, line_object: Callable, *args, **kwargs) -> Any:
        """[summary].

        Args:
            line_object (Callable): [description]
            args (tuple): [description]
            kwargs (dict): [description]

        Returns:
            Any: [description]
        """
        new_object = line_object(*args, **kwargs)
        self._session.add(new_object)

        return new_object

    def save(self):
        """Save data in database."""
        self._session.commit()



"""Run the CLI."""

from pathlib import Path

import click

from cakemix.commands import add, remove, run


@click.group()
def cli():
    """Init the CLI."""
    cakemix_data_dir = Path.home() / '.cakemix'
    cakemixes_dir = cakemix_data_dir / 'cakemixes'

    if not cakemix_data_dir.exists():
        cakemix_data_dir.mkdir()

    if not cakemixes_dir.exists():
        cakemixes_dir.mkdir()


cli.add_command(add.add)
cli.add_command(remove.remove)
cli.add_command(run.run)



"""Define the "add" command."""

import shutil
from pathlib import Path

import click
import toml

from cakemix.database import Cakemix, Database
from cakemix.output import run_task


def read_cakemix_src(database: Database, cakemix_src: Path):
    """[summary].

    Args:
        database (Database): [description]
        cakemix_src (Path): [description]

    Returns:
        Tuple[Any]: Error and cakemix.
    """
    if not Path(cakemix_src / 'structure.yaml').exists():
        return ('File \"structure.yaml\" not found',)

    if not Path(cakemix_src / 'settings.toml').exists():
        return ('File \"settings.toml\" not found',)

    cakemix = database.add(
        database.cakemix_object,
        database,
        structure=Path(cakemix_src / 'structure.yaml').read_text(),
        **toml.loads(Path(cakemix_src / 'settings.toml').read_text()),
    )

    if not Path(cakemix_src / 'arguments.toml').exists():
        return ('File \"arguments.toml\" not found',)

    arguments = toml.loads(Path(cakemix_src / 'arguments.toml').read_text())

    for argument_name, argument_options in arguments.items():
        cakemix.add_argument(name=argument_name, **argument_options)

    return ('', cakemix)


def read_paths(src_dir: Path, cakemix: Cakemix):
    """[summary].

    Args:
        src_dir (Path): [description]
        cakemix (Cakemix): [description]

    Returns:
        Tuple[any]: Error
    """
    src_len = len(str(src_dir)) + 1

    for path in sorted(src_dir.rglob('*')):
        if path.is_dir():
            cakemix.add_path(
                path=str(path)[src_len:], content_type='directory',
            )
        else:
            try:  # noqa: WPS229
                path.read_text()
                cakemix.add_path(
                    path=str(path)[src_len:], content_type='plain_text',
                )
            except UnicodeDecodeError:
                cakemix.add_path(
                    path=str(path)[
                        src_len:
                    ], content_type='not_plain_text',
                )

    return ('',)


def save_cakemix(src_dir: Path, cakemix: Cakemix, database: Database):
    """Save the cakemix in database and in .cakemix/cakemixes dir.

    Args:
        src_dir (Path): [description]
        cakemix (Cakemix): [description]
        database (Database): [description]

    Returns:
        Tuple[Any]: Error
    """
    shutil.make_archive(
        str(Path.home() / '.cakemix' / 'cakemixes' / cakemix.name),
        'zip',
        src_dir,
    )

    database.save()

    return ('',)


@click.command()
@click.argument('src')
def add(src: str):
    """Add a cakemix.

    Args:
        src (str): Cakemix location.
    """
    src_dir = Path(src)
    cakemix_src = src_dir / '.cakemixsrc'

    with Database() as database:
        cakemix = run_task(
            'Reading .cakemixsrc...',
            '.cakemixsrc',
            read_cakemix_src,
            database,
            cakemix_src,
        )

        run_task(
            'Processing files...', 'Files processed', read_paths, src_dir, cakemix,
        )

        run_task(
            'Saving cakemix...',
            'Cakemix saved',
            save_cakemix,
            src_dir,
            cakemix,
            database,
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cakemix', 'cakemix.commands']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.3.1,<6.0.0',
 'SQLAlchemy-Utils>=0.36.8,<0.37.0',
 'SQLAlchemy>=1.3.22,<2.0.0',
 'click>=7.1.2,<8.0.0',
 'questionary>=1.9.0,<2.0.0',
 'rich>=9.5.1,<10.0.0',
 'toml>=0.10.2,<0.11.0']

extras_require = \
{':python_version < "3.8"': ['importlib-metadata>=3.3.0,<4.0.0']}

entry_points = \
{'console_scripts': ['cakemix = cakemix.cli:cli']}

setup_kwargs = {
    'name': 'cakemix-python',
    'version': '0.2.0',
    'description': 'Cakemix is a tool for creating code template generators like create-react-app and npm-init. Avoiding you to waste time organizing a project.',
    'long_description': '==============\nCakemix :cake:\n==============\n\n|build| |codecov| |Gitpod ready to code|\n\nEnglish | Português_\n\nCakemix is a tool for creating code template generators like :code:`create-react-app` and :code:`npm init`. Avoiding you to waste time organizing a project.\n\n.. links\n\n.. _Português: ./locales/pt-br/README.rst\n\n.. images\n\n.. |build| image:: https://img.shields.io/github/workflow/status/vadolasi/cakemix/Python%20package\n   :alt: Tests status\n\n.. |codecov| image:: https://codecov.io/gh/vadolasi/cakemix/branch/master/graphs/badge.svg?branch=master\n   :alt: Code corverage status\n\n.. |Gitpod ready to code| image:: https://img.shields.io/badge/Gitpod-ready--to--code-blue?logo=gitpod\n   :alt: Open in Gitpod.io\n   :target: https://gitpod.io/#https://github.com/vadolasi/cakemix\n',
    'author': 'Vitor Daniel',
    'author_email': 'vitor036daniel@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/vadolasi/cakemix/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6.1,<=3.9.1',
}


setup(**setup_kwargs)
